#include "touch.h"
#include "settings.h"
#include <util/atomic.h>
#include <avr/interrupt.h>
#include <string.h>

#define REPEAT_MASK     (KEY_LEFT | KEY_RIGHT)
#define REPEAT_START    50                      // after 500ms
#define REPEAT_NEXT     20                      // every 200ms

#define ADC_PRESCALER(fmax) (F_CPU / fmax < 2  ? 1 : \
                             F_CPU / fmax < 4  ? 2 : \
                             F_CPU / fmax < 8  ? 3 : \
                             F_CPU / fmax < 16 ? 4 : \
                             F_CPU / fmax < 32 ? 5 : \
                             F_CPU / fmax < 64 ? 6 : 7)
#define ADC_SAMPLE_FREQ(pre) (F_CPU / 13.0 / ((pre & 0x07) ? 1<<(pre & 0x07) : 2))

#define ADCSRA_START (1<<ADEN | 1<<ADSC | 1<<ADIE | ADC_PRESCALER(100e3))       // 50..200kHz
#define ADMUX_GND (1<<REFS0 | 1<<MUX3 | 1<<MUX2 | 1<<MUX1 | 1<<MUX0)

touch_t touch[2];
uint8_t key_state;                              // debounced key state:
uint8_t key_press;                              // key press detect
uint8_t key_rpt;                                // key long press and repeat

void touch_init(void)
{
  memset(touch, -1, sizeof(touch));
  DIDR0 = 1<<TOUCHMUX1 | 1<<TOUCHMUX0;
  ADMUX = ADMUX_GND;
  ADCSRA = ADCSRA_START;
}

ISR(ADC_vect)
{
  static uint8_t touch_no;
  touch[touch_no].val += ADC - (touch[touch_no].val>>TOUCH_AVG);
  switch (touch_no++)
  {
    case 0:
      TOUCH_PORT0 = 1;                                  // charge next Touch Input
      TOUCH_DDR0 = 1;
      TOUCH_DDR1 = 0;
      TOUCH_PORT1 = 0;                                  // open Touch Input
      ADMUX = 1<<REFS0 | TOUCHMUX1;                     // S&H
      break;
    default:
      touch_no = 0;
      TOUCH_PORT1 = 1;                                  // charge next Touch Input
      TOUCH_DDR1 = 1;
      TOUCH_DDR0 = 0;
      TOUCH_PORT0 = 0;                                  // open Touch Input
      ADMUX = 1<<REFS0 | TOUCHMUX0;                     // S&H
      break;
  }
  ADCSRA = ADCSRA_START;
  ADMUX = ADMUX_GND;                                    // S&H GND after conversion complete
  static uint8_t cnt_10ms;
  if (++cnt_10ms >= (uint64_t)(.5 + ADC_SAMPLE_FREQ(ADCSRA_START) * 10e-3))
  {
    cnt_10ms = 0;
    static uint8_t ct0, ct1, rpt;
    uint8_t i = 0;
    touch_t* tp = &touch[0];
    for (uint8_t mask = 1; mask <= 2; mask<<= 1)
    {
      if (tp->val < tp->zero)
        tp->zero = tp->val;                             // store released value
      else
        if (tp->val - tp->zero > (THRESHOLD<<TOUCH_AVG))
          i |= mask;
      tp++;
    }
    i ^= key_state;                                     // key changed ?
    ct0 = ~( ct0 & i );                                 // reset or count ct0
    ct1 = ct0 ^ (ct1 & i);                              // reset or count ct1
    i &= ct0 & ct1;                                     // count until roll over ?
    key_state ^= i;                                     // then toggle debounced state
    key_press |= key_state & i;                         // 0->1: key press detect
    if( (key_state & REPEAT_MASK) == 0 )                // check repeat function
       rpt = REPEAT_START;                              // start delay
    if( --rpt == 0 ){
      rpt = REPEAT_NEXT;                                // repeat delay
      key_rpt |= key_state & REPEAT_MASK;
    }
  }
}

uint8_t get_key_press( uint8_t key_mask )
{
  ATOMIC_BLOCK(ATOMIC_FORCEON){			        // read and clear atomic
    key_mask &= key_press;			        // read key(s)
    key_press ^= key_mask;			        // clear key(s)
  }
  return key_mask;
}

uint8_t get_key_rpt( uint8_t key_mask )
{
  ATOMIC_BLOCK(ATOMIC_FORCEON){			        // read and clear atomic
    key_mask &= key_rpt;			        // read key(s)
    key_rpt ^= key_mask;			        // clear key(s)
  }
  return key_mask;
}

uint8_t get_key_short( uint8_t key_mask )
{
  cli();					        // read key state and key press atomic
  return get_key_press( ~key_state & key_mask );
}

uint8_t get_key_long( uint8_t key_mask )
{
  return get_key_press( get_key_rpt( key_mask ));
}
